/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "checkobjects.h"
#include "type1.h"
#include "type2.h"

#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>

#include <gwenhywfar/error.h>


#define AQDB_DBFILE_TEST_IDLIST_ENTRIES 256




int check_type1_object(AQDB_DB *db, const TYPE1 *t)
{
  const GWEN_DATE *dt1;
  GWEN_DATE *dt2;
  const GWEN_TIME *ti1;
  GWEN_TIME *ti2;
  char numbuf[256];
  AQDB_ID oid;
  const char *s;

  oid=Type1_GetId(t);

  s=Type1_GetStringField1(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField1 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField1", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type1_GetDateField1(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField1 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  s=Type1_GetStringField2(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField2 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField2", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  if (Type1_GetUint32Field1(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  ti1=Type1_GetTimestampField1(t);
  if (ti1==NULL) {
    fprintf(stderr, "ERROR: Empty timestampField1\n");
    return GWEN_ERROR_BAD_DATA;
  }
  ti2=GWEN_Time_new(1980, 1, (oid % 10)+1, 12, 0, 0, 1);
  if (GWEN_Time_Compare(ti1, ti2)) {
    GWEN_BUFFER *tbuf1;
    GWEN_BUFFER *tbuf2;

    tbuf1=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti1, "YYYYMMDD hh:mm:ss", tbuf1);

    tbuf2=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti2, "YYYYMMDD hh:mm:ss", tbuf2);

    fprintf(stderr, "ERROR: timestampField1 in object %llu does not match [%s != %s] [%d]\n",
            (unsigned long long) Type1_GetId(t),
            GWEN_Buffer_GetStart(tbuf1),
            GWEN_Buffer_GetStart(tbuf2),
            Type1_GetId(t) % 10);
    GWEN_Buffer_free(tbuf2);
    GWEN_Buffer_free(tbuf1);
    GWEN_Time_free(ti2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Time_free(ti2);


  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type1_GetDateField2(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField2 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  if (Type1_GetUint32Field2(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}



/* stringfield 1 and 2 swapped */
int check_type1_object_swapped(AQDB_DB *db, const TYPE1 *t)
{
  const GWEN_DATE *dt1;
  GWEN_DATE *dt2;
  const GWEN_TIME *ti1;
  GWEN_TIME *ti2;
  char numbuf[256];
  AQDB_ID oid;
  const char *s;

  oid=Type1_GetId(t);

  s=Type1_GetStringField2(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField2 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField1", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type1_GetDateField1(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField1 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  s=Type1_GetStringField1(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField1 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField2", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  if (Type1_GetUint32Field1(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  ti1=Type1_GetTimestampField1(t);
  if (ti1==NULL) {
    fprintf(stderr, "ERROR: Empty timestampField1\n");
    return GWEN_ERROR_BAD_DATA;
  }
  ti2=GWEN_Time_new(1980, 1, (oid % 10)+1, 12, 0, 0, 1);
  if (GWEN_Time_Compare(ti1, ti2)) {
    GWEN_BUFFER *tbuf1;
    GWEN_BUFFER *tbuf2;

    tbuf1=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti1, "YYYYMMDD hh:mm:ss", tbuf1);

    tbuf2=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti2, "YYYYMMDD hh:mm:ss", tbuf2);

    fprintf(stderr, "ERROR: timestampField1 in object %llu does not match [%s != %s] [%d]\n",
            (unsigned long long) Type1_GetId(t),
            GWEN_Buffer_GetStart(tbuf1),
            GWEN_Buffer_GetStart(tbuf2),
            Type1_GetId(t) % 10);
    GWEN_Buffer_free(tbuf2);
    GWEN_Buffer_free(tbuf1);
    GWEN_Time_free(ti2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Time_free(ti2);


  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type1_GetDateField2(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField2 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  if (Type1_GetUint32Field2(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}



int check_type2_object(AQDB_DB *db, const TYPE2 *t)
{
  const GWEN_DATE *dt1;
  GWEN_DATE *dt2;
  const GWEN_TIME *ti1;
  GWEN_TIME *ti2;
  char numbuf[256];
  AQDB_ID oid;
  const char *s;

  oid=Type2_GetId(t);

  s=Type2_GetStringField1(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField1 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField1", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type2_GetDateField1(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField1 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  s=Type2_GetStringField2(t);
  if (!(s && *s)) {
    fprintf(stderr, "ERROR: stringField2 is empty\n");
    return GWEN_ERROR_BAD_DATA;
  }
  snprintf(numbuf, sizeof(numbuf)-1, "%d-stringField2", oid);
  numbuf[sizeof(numbuf)-1]=0;
  if (strcasecmp(s, numbuf)) {
    fprintf(stderr, "ERROR: stringField2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  if (Type2_GetUint32Field1(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field1 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }


  ti1=Type2_GetTimestampField1(t);
  if (ti1==NULL) {
    fprintf(stderr, "ERROR: Empty timestampField1\n");
    return GWEN_ERROR_BAD_DATA;
  }
  ti2=GWEN_Time_new(1980, 1, (oid % 10)+1, 12, 0, 0, 1);
  if (GWEN_Time_Compare(ti1, ti2)) {
    GWEN_BUFFER *tbuf1;
    GWEN_BUFFER *tbuf2;

    tbuf1=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti1, "YYYYMMDD hh:mm:ss", tbuf1);

    tbuf2=GWEN_Buffer_new(0, 64, 0, 1);
    GWEN_Time_toString(ti2, "YYYYMMDD hh:mm:ss", tbuf2);

    fprintf(stderr, "ERROR: timestampField1 in object %llu does not match [%s != %s] [%d]\n",
            (unsigned long long) Type2_GetId(t),
            GWEN_Buffer_GetStart(tbuf1),
            GWEN_Buffer_GetStart(tbuf2),
            Type2_GetId(t) % 10);
    GWEN_Buffer_free(tbuf2);
    GWEN_Buffer_free(tbuf1);
    GWEN_Time_free(ti2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Time_free(ti2);


  /* date is 2000 / 01 / (objectId mod 20)+1 */
  dt1=Type2_GetDateField2(t);
  dt2=GWEN_Date_fromGregorian(2000, 1, (oid % 20)+1);
  if (GWEN_Date_Compare(dt1, dt2)) {
    fprintf(stderr, "ERROR: dateField2 does not match\n");
    GWEN_Date_free(dt2);
    return GWEN_ERROR_BAD_DATA;
  }
  GWEN_Date_free(dt2);


  if (Type2_GetUint32Field2(t)!=oid) {
    fprintf(stderr, "ERROR: uint32Field2 does not match\n");
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}




static void *GWENHYWFAR_CB read_type1_object_cb(AQDB_DB *db,
                                                AQDB_ID tableId,
                                                AQDB_OBJECT *o,
                                                void *data)
{
  TYPE1_LIST *tl;
  int rv;
  TYPE1 *t=NULL;

  tl=(TYPE1_LIST *) data;

  rv=Type1_fromObject(o, &t);
  if (rv<0) {
    fprintf(stderr, "Unable to read type1 from object %llu/%llu (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }
  Type1_SetDbId(t, AQDB_Object_GetObjectId(o));
  Type1_List_Add(t, tl);

  /* continue */
  return NULL;
}



int read_type1_objects(AQDB_DB *db, TYPE1_LIST *tl, AQDB_ID tableId, const char *queryExpr)
{
  int rv;
  void *result=NULL;

  rv=AQDB_DB_ForEveryQueryObject(db, tableId, 0, queryExpr, 0, read_type1_object_cb, (void *) tl, &result);
  if (rv<0) {
    fprintf(stderr, "ERROR: Error in AQDB_DB_ForEveryQueryObject() (%d)\n", rv);
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}



static void *GWENHYWFAR_CB read_type2_object_cb(AQDB_DB *db,
                                                AQDB_ID tableId,
                                                AQDB_OBJECT *o,
                                                void *data)
{
  TYPE2_LIST *tl;
  int rv;
  TYPE2 *t=NULL;

  tl=(TYPE2_LIST *) data;

  rv=Type2_fromObject(o, &t);
  if (rv<0) {
    fprintf(stderr, "Unable to read type2 from object %llu/%llu (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }
  Type2_List_Add(t, tl);

  /* continue */
  return NULL;
}



int read_type2_objects(AQDB_DB *db, TYPE2_LIST *tl, AQDB_ID tableId, const char *queryExpr)
{
  int rv;
  void *result=NULL;

  rv=AQDB_DB_ForEveryQueryObject(db, tableId, 0, queryExpr, 0, read_type2_object_cb, (void *) tl, &result);
  if (rv<0) {
    fprintf(stderr, "ERROR: Error in AQDB_DB_ForEveryQueryObject() (%d)\n", rv);
    return GWEN_ERROR_BAD_DATA;
  }

  return 0;
}




static void *GWENHYWFAR_CB read_type1_object_into_idlist_cb(AQDB_DB *db,
                                                            AQDB_ID tableId,
                                                            AQDB_OBJECT *o,
                                                            void *data)
{
  GWEN_IDLIST64 *idl;
  int rv;
  TYPE1 *t=NULL;
  AQDB_ID currentId;

  idl=(GWEN_IDLIST64 *) data;

  rv=Type1_fromObject(o, &t);
  if (rv<0) {
    fprintf(stderr, "Unable to read type1 from object %llu/%llu (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }

  rv=check_type1_object(db, t);
  if (rv<0) {
    fprintf(stderr, "Error checking type1 object %llu/%llu (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }

  currentId=Type1_GetId(t);
  if (currentId==0) {
    fprintf(stderr, "Type1 object from object %llu/%llu has no id (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }

  rv=GWEN_IdList64_AddId(idl, currentId);
  if (rv<0) {
    fprintf(stderr, "Unable to add id for type1 from object %llu/%llu (%d)\n",
            (unsigned long long) AQDB_Object_GetTableId(o),
            (unsigned long long) AQDB_Object_GetObjectId(o),
            rv);
    return (void *) o;
  }


  Type1_free(t);

  /* continue */
  return NULL;
}



GWEN_IDLIST64 *read_type1_objects_into_idlist(AQDB_DB *db, AQDB_ID tableId, const char *queryExpr)
{
  GWEN_IDLIST64 *idList1;
  int rv;
  void *result=NULL;

  idList1=GWEN_IdList64_newWithSteps(AQDB_DBFILE_TEST_IDLIST_ENTRIES);
  rv=AQDB_DB_ForEveryQueryObject(db, tableId, 0, queryExpr, 0, read_type1_object_into_idlist_cb, (void *) idList1,
                                 &result);
  if (rv<0) {
    fprintf(stderr, "ERROR: Error in AQDB_DB_ForEveryQueryObject() (%d)\n", rv);
    GWEN_IdList64_free(idList1);
    return NULL;
  }


  return idList1;
}



