/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2020 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "createdb.h"
#include "checkobjects.h"
#include "createobjects.h"
#include "type1.h"
#include "type2.h"

#include <gwenhywfar/error.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/testframework.h>



#define RUN1_TYPE1_NUM     250
#define RUN1_TYPE1_STARTID 10000

#define RUN1_TYPE2_NUM     200
#define RUN1_TYPE2_STARTID 20000

#define RUN2_TYPE1_NUM     250
#define RUN2_TYPE1_STARTID 30000

#define RUN_LOOP_NUM       250
#define RUN_LOOP_ADD       4
#define RUN_LOOP_DEL       3




static AQDB_DB *_createDataBaseWithObjects(const char *url,
                                           const char *dirname,
                                           const char *fname,
                                           AQDB_ID *pTableId1,
                                           AQDB_ID *pTableId2);

static GWEN_IDLIST64 *_removeSomeAndCheck(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numToRemove);
static GWEN_IDLIST64 *_removeSome(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numToRemove);
static int _testDatabaseAgainstIdList(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId);
/*static int _testMembersFromList1InList2(GWEN_IDLIST64 *idList1, GWEN_IDLIST64 *idList2);*/
static int _compareList1AgainstList2(GWEN_IDLIST64 *idList1, GWEN_IDLIST64 *idList2);
static AQDB_ID _getDatabaseId(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId);
static int _deleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId);
static GWEN_IDLIST64 *_addSomeAndCheck(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numberOfObjects,
                                       AQDB_ID startId);
static GWEN_IDLIST64 *_addSome(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numberOfObjects,
                               AQDB_ID startId);




int test_remove_objects(GWEN_TEST_MODULE *mod)
{
  AQDB_DB *db;
  AQDB_ID tableId1=0;
  AQDB_ID tableId2=0;
  GWEN_IDLIST64 *idList1=NULL;
  AQDB_ID nextObjectId;
  const char *url;
  const char *dirname;
  const char *fname;
  int rv;
  int i;

  /* create database */
  url=GWEN_Test_Module_GetCharParam(mod, "url", NULL);
  dirname=GWEN_Test_Module_GetCharParam(mod, "dirName", NULL);
  fname=GWEN_Test_Module_GetCharParam(mod, "fname", NULL);

  destroy_database(url, dirname, fname);
  db=_createDataBaseWithObjects(url, dirname, fname, &tableId1, &tableId2);
  if (db==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "ERROR: Could not create database.");
    return GWEN_ERROR_GENERIC;
  }

  /* read objects into idList */
  idList1=read_type1_objects_into_idlist(db, tableId1, NULL);
  if (idList1==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "ERROR: Problem reading and checking objects");
    AQDB_DB_free(db);
    return GWEN_ERROR_GENERIC;
  }

  rv=GWEN_IdList64_Sort(idList1);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* pretest: check whether idlist stuff really works */
  rv=_testDatabaseAgainstIdList(db, idList1, tableId1);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* loop: remove objects, add objects */
  nextObjectId=RUN2_TYPE1_STARTID;
  for (i=0; i<RUN_LOOP_NUM; i++) {
    GWEN_IDLIST64 *idListStage1;
    GWEN_IDLIST64 *idListStage2;

    idListStage1=_removeSomeAndCheck(db, idList1, tableId1, RUN_LOOP_DEL);
    if (idListStage1==NULL) {
      DBG_ERROR(0, "ERROR: _removeSomeAndCheck");
      AQDB_DB_free(db);
      return rv;
    }

    idListStage2=_addSomeAndCheck(db, idListStage1, tableId1, RUN_LOOP_ADD, nextObjectId);
    if (idListStage2==NULL) {
      DBG_ERROR(0, "ERROR: _addSomeAndCheck");
      AQDB_DB_free(db);
      return rv;
    }

    GWEN_IdList64_free(idListStage1);
    GWEN_IdList64_free(idList1);
    idList1=idListStage2;

    rv=GWEN_IdList64_Sort(idList1);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DB_free(db);
      return rv;
    }

    nextObjectId+=RUN_LOOP_ADD;
  }

  rv=AQDB_DB_Close(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return 2;
  }

  AQDB_DB_free(db);

  return 0;
}



AQDB_DB *_createDataBaseWithObjects(const char *url,
                                    const char *dirname,
                                    const char *fname,
                                    AQDB_ID *pTableId1,
                                    AQDB_ID *pTableId2)
{
  AQDB_DB *db=NULL;
  AQDB_ID tableId1;
  AQDB_ID tableId2;
  int rv;

  /* create database */
  rv=create_database(url, dirname, fname);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }


  rv=AQDB_DbFactory(url, &db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return NULL;
  }

  AQDB_DB_SetCache(db, 1000, 10*1024*1024);

  /* open database */
  rv=AQDB_DB_Open(db, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* open table 1 */
  tableId1=0;
  rv=AQDB_DB_OpenTable(db, "type1",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId1);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* open table 2 */
  tableId2=0;
  rv=AQDB_DB_OpenTable(db, "type2",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId2);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* begin edit session */
  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* create type 1 objects */
  rv=create_type1_objects(db, tableId1, RUN1_TYPE1_NUM, RUN1_TYPE1_STARTID); /* create x objects, starting with 2000 */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* create type 2 objects */
  rv=create_type2_objects(db, tableId2, RUN1_TYPE2_NUM, RUN1_TYPE2_STARTID); /* create x objects, starting with 7000 */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  /* done editing */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return NULL;
  }

  *pTableId1=tableId1;
  *pTableId2=tableId2;

  return db;
}


GWEN_IDLIST64 *_addSomeAndCheck(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numberOfObjects,
                                AQDB_ID startId)
{
  int rv;
  GWEN_IDLIST64 *idListResult;

  /* begin edit session */
  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    return NULL;
  }

  idListResult=_addSome(db, idListSource, tableId, numberOfObjects, startId);
  if (idListResult==NULL) {
    DBG_ERROR(0, "here");
    AQDB_DB_EndEdit(db, 1);
    return NULL;
  }

  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    AQDB_DB_EndEdit(db, 1);
    return NULL;
  }

  rv=GWEN_IdList64_Sort(idListResult);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_IdList64_free(idListResult);
    return NULL;
  }

  rv=_testDatabaseAgainstIdList(db, idListResult, tableId);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_IdList64_free(idListResult);
    return NULL;
  }

  return idListResult;
}



GWEN_IDLIST64 *_addSome(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numberOfObjects, AQDB_ID startId)
{
  int rv;
  GWEN_IDLIST64 *idListResult;
  int i;
  AQDB_ID id;

  /* create type 1 objects */
  rv=create_type1_objects(db, tableId, numberOfObjects, startId);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    AQDB_DB_EndEdit(db, 1);
    return NULL;
  }

  idListResult=GWEN_IdList64_dup(idListSource);
  id=startId;
  for (i=0; i<numberOfObjects; i++) {
    GWEN_IdList64_AddId(idListResult, id);
    id++;
  }

  return idListResult;
}



GWEN_IDLIST64 *_removeSomeAndCheck(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numToRemove)
{
  int rv;
  GWEN_IDLIST64 *idListResult;

  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    return NULL;
  }

  idListResult=_removeSome(db, idListSource, tableId, numToRemove);
  if (idListResult==NULL) {
    DBG_ERROR(0, "here (%d)", rv);
    AQDB_DB_EndEdit(db, 1);
    return NULL;
  }

  /* done editing */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_IdList64_free(idListResult);
    return NULL;
  }

  rv=GWEN_IdList64_Sort(idListResult);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_IdList64_free(idListResult);
    return NULL;
  }

  rv=_testDatabaseAgainstIdList(db, idListResult, tableId);
  if (rv<0) {
    DBG_ERROR(0, "here (%d)", rv);
    GWEN_IdList64_free(idListResult);
    return NULL;
  }

  return idListResult;
}



GWEN_IDLIST64 *_removeSome(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId, int numToRemove)
{
  int i;
  uint64_t numOfEntries;
  GWEN_IDLIST64 *idListDest;

  idListDest=GWEN_IdList64_dup(idListSource);
  if (idListDest==NULL) {
    DBG_ERROR(0, "here");
    return NULL;
  }
  numOfEntries=GWEN_IdList64_GetEntryCount(idListDest);

  for (i=0; i<numToRemove; i++) {
    uint64_t id=0;
    int rv;

    /* find a non-empty id */
    while (id==0) {
      long int r;

      r=random() % numOfEntries;
      id=GWEN_IdList64_GetIdAt(idListDest, r);
    }

    /* delete object */
    rv=_deleteObject(db, tableId, id);
    if (rv<0) {
      DBG_ERROR(0, "Unable to delete object %lu (%d)", (unsigned long) id, rv);
      GWEN_IdList64_free(idListDest);
      return NULL;
    }
    GWEN_IdList64_DelId(idListDest, id);
    numOfEntries--;
  }

  return idListDest;
}



int _deleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId)
{
  int rv;
  AQDB_ID idForDatabase;

  idForDatabase=_getDatabaseId(db, tableId, objectId);
  if (idForDatabase==0) {
    DBG_ERROR(0, "Could not get object id for database");
    return GWEN_ERROR_GENERIC;
  }

  rv=AQDB_DB_DeleteObject(db, tableId, idForDatabase);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    return rv;
  }

  return 0;
}



AQDB_ID _getDatabaseId(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId)
{
  int rv;
  char queryBuf[256];
  AQDB_ID qid=0;
  AQDB_ID result=0;
  int len;

  rv=snprintf(queryBuf, sizeof(queryBuf), "$id==%lu", (unsigned long) objectId);
  if (rv>=sizeof(queryBuf)) {
    DBG_INFO(0, "Buffer too small (%d)", rv);
    return 0;
  }

  rv=AQDB_DB_QuerySubmit(db, tableId, 0, queryBuf, AQDB_QUERY_FLAGS_STOPATFIRST, &qid);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    return 0;
  }

  /* get single first id */
  len=1;
  rv=AQDB_DB_QueryGetFirst(db, qid, &result, &len);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    AQDB_DB_QueryClose(db, qid);
    return 0;
  }
  AQDB_DB_QueryClose(db, qid);

  if (len!=1) {
    DBG_ERROR(0, "Query did not return an object id");
    return GWEN_ERROR_GENERIC;
  }

  return result;
}



int _testDatabaseAgainstIdList(AQDB_DB *db, GWEN_IDLIST64 *idListSource, AQDB_ID tableId1)
{
  GWEN_IDLIST64 *idList1;
  int rv;

  idList1=read_type1_objects_into_idlist(db, tableId1, NULL);
  if (idList1==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "ERROR: Problem reading and checking objects\n");
    GWEN_IdList64_free(idList1);
    return GWEN_ERROR_GENERIC;
  }

  rv=GWEN_IdList64_Sort(idList1);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    GWEN_IdList64_free(idList1);
    return rv;
  }

  rv=_compareList1AgainstList2(idList1, idListSource);
  if (rv<0) {
    DBG_INFO(0, "here (%d)", rv);
    GWEN_IdList64_free(idList1);
    return rv;
  }

  GWEN_IdList64_free(idList1);
  return 0;
}



#if 0
int _testMembersFromList1InList2(GWEN_IDLIST64 *idList1, GWEN_IDLIST64 *idList2)
{
  GWEN_IDLIST64_ITERATOR *iter;
  uint64_t id;

  iter=GWEN_IdList64_Iterator_new(idList1);
  id=GWEN_IdList64_Iterator_GetFirstId(iter);
  while (id) {
    if (GWEN_IdList64_HasId(idList2, id)!=0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table 2 does not have id %lu", (unsigned long) id);
      return GWEN_ERROR_GENERIC;
    }
    id=GWEN_IdList64_Iterator_GetNextId(iter);
  }
  GWEN_IdList64_Iterator_free(iter);

  return 0;
}
#endif



int _compareList1AgainstList2(GWEN_IDLIST64 *idList1, GWEN_IDLIST64 *idList2)
{
  GWEN_IDLIST64_ITERATOR *iter1;
  GWEN_IDLIST64_ITERATOR *iter2;
  uint64_t id1;
  uint64_t id2;
  uint64_t index=0;

  iter1=GWEN_IdList64_Iterator_new(idList1);
  iter2=GWEN_IdList64_Iterator_new(idList2);
  id1=GWEN_IdList64_Iterator_GetFirstId(iter1);
  id2=GWEN_IdList64_Iterator_GetFirstId(iter2);
  while (id1>0 && id2>0) {
    if (id1 != id2) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Tables differ at position %lu (%lu != %lu)",
                (unsigned long) index,
                (unsigned long) id1,
                (unsigned long) id2);
      GWEN_IdList64_Iterator_free(iter2);
      GWEN_IdList64_Iterator_free(iter1);
      return GWEN_ERROR_GENERIC;
    }
    index++;
    id1=GWEN_IdList64_Iterator_GetNextId(iter1);
    id2=GWEN_IdList64_Iterator_GetNextId(iter2);
  }
  if (GWEN_IdList64_Iterator_GetNextId(iter1)>0 || GWEN_IdList64_Iterator_GetNextId(iter2)>0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Number of elements in lists differs");
    GWEN_IdList64_Iterator_free(iter2);
    GWEN_IdList64_Iterator_free(iter1);
    return GWEN_ERROR_GENERIC;
  }
  GWEN_IdList64_Iterator_free(iter2);
  GWEN_IdList64_Iterator_free(iter1);

  return 0;
}


