/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "createdb.h"
#include "checkobjects.h"
#include "createobjects.h"
#include "type1.h"
#include "type2.h"

#include <gwenhywfar/error.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/testframework.h>


#define RUN1_TYPE1_NUM     1000
#define RUN1_TYPE1_STARTID 10000

#define RUN1_TYPE2_NUM     800
#define RUN1_TYPE2_STARTID 20000


#define RUN2_TYPE1_NUM     700
#define RUN2_TYPE1_STARTID 30000

#define RUN2_TYPE2_NUM     900
#define RUN2_TYPE2_STARTID 40000




int test_create_modify2(GWEN_TEST_MODULE *mod)
{
  AQDB_DB *db=NULL;
  AQDB_ID tableId1;
  AQDB_ID tableId2;
  const char *url;
  const char *dirname;
  const char *fname;
  int rv;

  /* create database */
  url=GWEN_Test_Module_GetCharParam(mod, "url", NULL);
  dirname=GWEN_Test_Module_GetCharParam(mod, "dirName", NULL);
  fname=GWEN_Test_Module_GetCharParam(mod, "fname", NULL);

  destroy_database(url, dirname, fname);

  rv=create_database(url, dirname, fname);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }


  rv=AQDB_DbFactory(url, &db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  AQDB_DB_SetCache(db, 1000, 10*1024*1024);

  /* open database */
  rv=AQDB_DB_Open(db, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* open table 1 */
  tableId1=0;
  rv=AQDB_DB_OpenTable(db, "type1",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId1);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* open table 2 */
  tableId2=0;
  rv=AQDB_DB_OpenTable(db, "type2",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId2);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }


  /* ------------------------------------------------------------------------------------- run 1 ----*/

  /* begin edit session */
  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* create type 1 objects */
  rv=create_type1_objects(db, tableId1, RUN1_TYPE1_NUM, RUN1_TYPE1_STARTID); /* create x objects, starting with 2000 */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* create type 2 objects */
  rv=create_type2_objects(db, tableId2, RUN1_TYPE2_NUM, RUN1_TYPE2_STARTID); /* create x objects, starting with 7000 */
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* done editing */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }


  /* ------------------------------------------------------------------------------------- run 2 ----*/

  /* begin 2nd edit session */
  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }


  /* swap stringfields */
  if (1) {
    TYPE1_LIST *tl;
    TYPE1 *t;
    GWEN_BUFFER *tbuf;

    tl=Type1_List_new();
    rv=read_type1_objects(db, tl, tableId1, NULL);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DB_free(db);
      return rv;
    }

    /* swap stringfields 1 and 2 */
    tbuf=GWEN_Buffer_new(0, 64, 0, 1);
    t=Type1_List_First(tl);
    while (t) {
      const char *s1;
      const char *s2;
      AQDB_OBJECT *o=NULL;
      AQDB_ID currentDbId;

      currentDbId=Type1_GetDbId(t);
      s1=Type1_GetStringField1(t);
      s2=Type1_GetStringField2(t);
      GWEN_Buffer_AppendString(tbuf, s2);
      Type1_SetStringField2(t, s1);
      Type1_SetStringField1(t, GWEN_Buffer_GetStart(tbuf));
      GWEN_Buffer_Reset(tbuf);

      /* store object */
      rv=Type1_CreateObject(db, tableId1, currentDbId, &o);
      if (rv<0) {
        fprintf(stderr, "ERROR: Unable to create object (%d)\n", rv);
        return rv;
      }
      rv=Type1_toObject(t, o);
      if (rv<0) {
        DBG_ERROR(0, "ERROR: Unable to transform type to object (%d)", rv);
        AQDB_Object_free(o);
        return rv;
      }

      rv=AQDB_DB_WriteObject(db, o);
      if (rv<0) {
        DBG_ERROR(0, "ERROR: Unable to write object (%d)", rv);
        AQDB_Object_free(o);
        return rv;
      }
      AQDB_Object_free(o);


      t=Type1_List_Next(t);
    }
    GWEN_Buffer_free(tbuf);

    Type1_List_free(tl);
  }

  if (1) {
    TYPE1_LIST *tl;
    TYPE1 *t;

    tl=Type1_List_new();
    rv=read_type1_objects(db, tl, tableId1, NULL);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DB_free(db);
      return rv;
    }

    /* sort list */
    Type1_List_SortById(tl, 1);

    t=Type1_List_First(tl);
    while (t) {
      rv=check_type1_object_swapped(db, t);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_DB_free(db);
        return rv;
      }

      t=Type1_List_Next(t);
    }
    Type1_List_free(tl);
  }

  /* done editing */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return rv;
  }


  rv=AQDB_DB_Close(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_free(db);
    return 2;
  }

  AQDB_DB_free(db);

  return 0;
}


